"""nova_utils.data

Author:
    Dominik Schiller <dominik.schiller@uni-a.de>
Date:
    06th September 2023

The Nova-Utils Data package defines data types for streams and annotations as well as providing respective data handler to load and save data from and to various backends.
"""